import asyncio
import aiomysql


loop = asyncio.get_event_loop()


@asyncio.coroutine
def test_example():
        pool = yield from aiomysql.create_pool(host='127.0.0.1', port=3306,
                                               user='root', password='',
                                               db='mysql', loop=loop)
        with (yield from pool) as conn:
            cur = yield from conn.cursor()
            yield from cur.execute("SELECT 10")
            # print(cur.description)
            (r,) = yield from cur.fetchone()
            assert r == 10
        pool.close()
        yield from pool.wait_closed()


loop.run_until_complete(test_example())



import copy
import asyncio

import aiomysql.cursors
from tests import base
from tests._testutils import run_until_complete


class TestDeserializeCursor(base.AIOPyMySQLTestCase):
    bob = ("bob", 21, {"k1": "pretty", "k2": [18, 25]})
    jim = ("jim", 56, {"k1": "rich", "k2": [20, 60]})
    fred = ("fred", 100, {"k1": "longevity", "k2": [100, 160]})
    havejson = True

    cursor_type = aiomysql.cursors.DeserializationCursor

    def setUp(self):
        super(TestDeserializeCursor, self).setUp()
        self.conn = conn = self.connections[0]

        @asyncio.coroutine
        def prepare():
            c = yield from conn.cursor(self.cursor_type)

            # create a table ane some data to query
            yield from c.execute("drop table if exists deserialize_cursor")
            yield from c.execute("select VERSION()")
            v = yield from c.fetchone()
            version, *db_type = v[0].split('-', 1)
            version = float(".".join(version.split('.', 2)[:2]))
            ismariadb = db_type and 'mariadb' in db_type[0].lower()
            if ismariadb or version < 5.7:
                yield from c.execute(
                    """CREATE TABLE deserialize_cursor
                     (name char(20), age int , claim text)""")
                self.havejson = False
            else:
                yield from c.execute(
                    """CREATE TABLE deserialize_cursor
                     (name char(20), age int , claim json)""")
            data = [("bob", 21, '{"k1": "pretty", "k2": [18, 25]}'),
                    ("jim", 56, '{"k1": "rich", "k2": [20, 60]}'),
                    ("fred", 100, '{"k1": "longevity", "k2": [100, 160]}')]
            yield from c.executemany("insert into deserialize_cursor values "
                                     "(%s,%s,%s)",
                                     data)

        self.loop.run_until_complete(prepare())

    def tearDown(self):
        @asyncio.coroutine
        def shutdown():
            c = yield from self.conn.cursor()
            yield from c.execute("drop table deserialize_cursor;")

        self.loop.run_until_complete(shutdown())
        super(TestDeserializeCursor, self).tearDown()

    @run_until_complete
    def test_deserialize_cursor(self):
        if not self.havejson:
            return
        bob, jim, fred = copy.deepcopy(self.bob), copy.deepcopy(
            self.jim), copy.deepcopy(self.fred)
        # all assert test compare to the structure as would come
        # out from MySQLdb
        conn = self.conn
        c = yield from conn.cursor(self.cursor_type)

        # pull back the single row dict for bob and check
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchone()
        self.assertEqual(bob, r, "fetchone via DeserializeCursor failed")
        # same again, but via fetchall => tuple)
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchall()
        self.assertEqual([bob], r,
                         "fetch a 1 row result via fetchall failed via "
                         "DeserializeCursor")
        # get all 3 row via fetchall
        yield from c.execute("SELECT * from deserialize_cursor")
        r = yield from c.fetchall()
        self.assertEqual([bob, jim, fred], r,
                         "fetchall failed via DictCursor")

        # get all 2 row via fetchmany
        yield from c.execute("SELECT * from deserialize_cursor")
        r = yield from c.fetchmany(2)
        self.assertEqual([bob, jim], r, "fetchmany failed via DictCursor")
        yield from c.execute('commit')

    @run_until_complete
    def test_deserialize_cursor_low_version(self):
        if self.havejson:
            return
        bob = ("bob", 21, '{"k1": "pretty", "k2": [18, 25]}')
        jim = ("jim", 56, '{"k1": "rich", "k2": [20, 60]}')
        fred = ("fred", 100, '{"k1": "longevity", "k2": [100, 160]}')
        # all assert test compare to the structure as would come
        # out from MySQLdb
        conn = self.conn
        c = yield from conn.cursor(self.cursor_type)

        # pull back the single row dict for bob and check
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchone()
        self.assertEqual(bob, r, "fetchone via DeserializeCursor failed")
        # same again, but via fetchall => tuple)
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchall()
        self.assertEqual([bob], r,
                         "fetch a 1 row result via fetchall failed via "
                         "DeserializeCursor")
        # get all 3 row via fetchall
        yield from c.execute("SELECT * from deserialize_cursor")
        r = yield from c.fetchall()
        self.assertEqual([bob, jim, fred], r,
                         "fetchall failed via DictCursor")

        # get all 2 row via fetchmany
        yield from c.execute("SELECT * from deserialize_cursor")
        r = yield from c.fetchmany(2)
        self.assertEqual([bob, jim], r, "fetchmany failed via DictCursor")
        yield from c.execute('commit')

    @run_until_complete
    def test_deserializedictcursor(self):
        if not self.havejson:
            return
        bob = {'name': 'bob', 'age': 21,
               'claim': {"k1": "pretty", "k2": [18, 25]}}
        conn = self.conn
        c = yield from conn.cursor(aiomysql.cursors.DeserializationCursor,
                                   aiomysql.cursors.DictCursor)
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchall()
        self.assertEqual([bob], r,
                         "fetch a 1 row result via fetchall failed via "
                         "DeserializationCursor")

    @run_until_complete
    def test_ssdeserializecursor(self):
        if not self.havejson:
            return
        conn = self.conn
        c = yield from conn.cursor(aiomysql.cursors.SSCursor,
                                   aiomysql.cursors.DeserializationCursor)
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchall()
        self.assertEqual([self.bob], r,
                         "fetch a 1 row result via fetchall failed via "
                         "DeserializationCursor")

    @run_until_complete
    def test_ssdeserializedictcursor(self):
        if not self.havejson:
            return
        bob = {'name': 'bob', 'age': 21,
               'claim': {"k1": "pretty", "k2": [18, 25]}}
        conn = self.conn
        c = yield from conn.cursor(aiomysql.cursors.SSCursor,
                                   aiomysql.cursors.DeserializationCursor,
                                   aiomysql.cursors.DictCursor)
        yield from c.execute("SELECT * from deserialize_cursor "
                             "where name='bob'")
        r = yield from c.fetchall()
        self.assertEqual([bob], r,
                         "fetch a 1 row result via fetchall failed via "
                         "DeserializationCursor")



import asyncio
from aiomysql import connect, sa
from enum import IntEnum

import os
import unittest
from unittest import mock

from sqlalchemy import MetaData, Table, Column, Integer, TypeDecorator


class UserDefinedEnum(IntEnum):
    Value1 = 111
    Value2 = 222


class IntEnumField(TypeDecorator):
    impl = Integer

    def __init__(self, enum_class, *arg, **kw):
        TypeDecorator.__init__(self, *arg, **kw)
        self.enum_class = enum_class

    def process_bind_param(self, value, dialect):
        """ From python to DB """
        if value is None:
            return None
        elif not isinstance(value, self.enum_class):
            return self.enum_class(value).value
        else:
            return value.value

    def process_result_value(self, value, dialect):
        """ From DB to Python """
        if value is None:
            return None

        return self.enum_class(value)


meta = MetaData()
tbl = Table('sa_test_type_tbl', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('val', IntEnumField(enum_class=UserDefinedEnum)))


class TestSATypes(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.host = os.environ.get('MYSQL_HOST', 'localhost')
        self.port = int(os.environ.get('MYSQL_PORT', 3306))
        self.user = os.environ.get('MYSQL_USER', 'root')
        self.db = os.environ.get('MYSQL_DB', 'test_pymysql')
        self.password = os.environ.get('MYSQL_PASSWORD', '')

    def tearDown(self):
        self.loop.close()

    async def connect(self, **kwargs):
        conn = await connect(db=self.db,
                             user=self.user,
                             password=self.password,
                             host=self.host,
                             loop=self.loop,
                             port=self.port,
                             **kwargs)
        await conn.autocommit(True)
        cur = await conn.cursor()
        await cur.execute("DROP TABLE IF EXISTS sa_test_type_tbl")
        await cur.execute("CREATE TABLE sa_test_type_tbl "
                          "(id serial, val bigint)")
        await cur._connection.commit()
        engine = mock.Mock()
        engine.dialect = sa.engine._dialect
        return sa.SAConnection(conn, engine)

    def test_values(self):
        async def go():
            conn = await self.connect()

            await conn.execute(tbl.insert().values(
                val=UserDefinedEnum.Value1)
            )
            result = await conn.execute(tbl.select().where(
                tbl.c.val == UserDefinedEnum.Value1)
            )
            data = await result.fetchone()
            self.assertEqual(
                data['val'], UserDefinedEnum.Value1
            )

        self.loop.run_until_complete(go())



import asyncio
from aiomysql import sa
from sqlalchemy import bindparam

import os
import unittest

from sqlalchemy import MetaData, Table, Column, Integer, String

meta = MetaData()
tbl = Table('sa_tbl_cache_test', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('val', String(255)))


class TestCompiledCache(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.host = os.environ.get('MYSQL_HOST', 'localhost')
        self.port = int(os.environ.get('MYSQL_PORT', 3306))
        self.user = os.environ.get('MYSQL_USER', 'root')
        self.db = os.environ.get('MYSQL_DB', 'test_pymysql')
        self.password = os.environ.get('MYSQL_PASSWORD', '')
        self.engine = self.loop.run_until_complete(self.make_engine())
        self.loop.run_until_complete(self.start())

    def tearDown(self):
        self.engine.terminate()
        self.loop.run_until_complete(self.engine.wait_closed())
        self.loop.close()

    async def make_engine(self, **kwargs):
            return (await sa.create_engine(db=self.db,
                                           user=self.user,
                                           password=self.password,
                                           host=self.host,
                                           port=self.port,
                                           loop=self.loop,
                                           minsize=10,
                                           **kwargs))

    async def start(self):
        async with self.engine.acquire() as conn:
            tx = await conn.begin()
            await conn.execute("DROP TABLE IF EXISTS "
                               "sa_tbl_cache_test")
            await conn.execute("CREATE TABLE sa_tbl_cache_test"
                               "(id serial, val varchar(255))")
            await conn.execute(tbl.insert().values(val='some_val_1'))
            await conn.execute(tbl.insert().values(val='some_val_2'))
            await conn.execute(tbl.insert().values(val='some_val_3'))
            await tx.commit()

    def test_cache(self):
        async def go():
            cache = dict()
            engine = await self.make_engine(compiled_cache=cache)
            async with engine.acquire() as conn:
                # check select with params not added to cache
                q = tbl.select().where(tbl.c.val == 'some_val_1')
                cursor = await conn.execute(q)
                row = await cursor.fetchone()
                self.assertEqual('some_val_1', row.val)
                self.assertEqual(0, len(cache))

                # check select with bound params added to cache
                select_by_val = tbl.select().where(
                    tbl.c.val == bindparam('value')
                )
                cursor = await conn.execute(
                    select_by_val, {'value': 'some_val_3'}
                )
                row = await cursor.fetchone()
                self.assertEqual('some_val_3', row.val)
                self.assertEqual(1, len(cache))

                cursor = await conn.execute(
                    select_by_val, value='some_val_2'
                )
                row = await cursor.fetchone()
                self.assertEqual('some_val_2', row.val)
                self.assertEqual(1, len(cache))

                select_all = tbl.select()
                cursor = await conn.execute(select_all)
                rows = await cursor.fetchall()
                self.assertEqual(3, len(rows))
                self.assertEqual(2, len(cache))

                # check insert with bound params not added to cache
                await conn.execute(tbl.insert().values(val='some_val_4'))
                self.assertEqual(2, len(cache))

                # check insert with bound params added to cache
                q = tbl.insert().values(val=bindparam('value'))
                await conn.execute(q, value='some_val_5')
                self.assertEqual(3, len(cache))

                await conn.execute(q, value='some_val_6')
                self.assertEqual(3, len(cache))

                await conn.execute(q, {'value': 'some_val_7'})
                self.assertEqual(3, len(cache))

                cursor = await conn.execute(select_all)
                rows = await cursor.fetchall()
                self.assertEqual(7, len(rows))
                self.assertEqual(3, len(cache))

                # check update with params not added to cache
                q = tbl.update().where(
                    tbl.c.val == 'some_val_1'
                ).values(val='updated_val_1')
                await conn.execute(q)
                self.assertEqual(3, len(cache))
                cursor = await conn.execute(
                    select_by_val, value='updated_val_1'
                )
                row = await cursor.fetchone()
                self.assertEqual('updated_val_1', row.val)

                # check update with bound params added to cache
                q = tbl.update().where(
                    tbl.c.val == bindparam('value')
                ).values(val=bindparam('update'))
                await conn.execute(
                    q, value='some_val_2', update='updated_val_2'
                )
                self.assertEqual(4, len(cache))
                cursor = await conn.execute(
                    select_by_val, value='updated_val_2'
                )
                row = await cursor.fetchone()
                self.assertEqual('updated_val_2', row.val)

        self.loop.run_until_complete(go())



import copy
from aiomysql import create_pool

import pytest


# You could parameterise these tests with this, but then pytest
# does some funky stuff and spins up and tears down containers
# per function call.  Remember it would be
# mysql_versions * event_loops * 4 auth tests ~= 3*2*4 ~= 24 tests

# As the MySQL daemon restarts at least 3 times in the container
# before it becomes stable, there's a sleep(10) so that's
# around a 4min wait time.

# @pytest.mark.parametrize("user,password,plugin", [
#     ("nopass_sha256", None, 'sha256_password'),
#     ("user_sha256", 'pass_sha256', 'sha256_password'),
#     ("nopass_caching_sha2", None, 'caching_sha2_password'),
#     ("user_caching_sha2", 'pass_caching_sha2', 'caching_sha2_password'),
# ])


@pytest.mark.mysql_verison('8.0')
@pytest.mark.run_loop
async def test_sha256_nopw(mysql_server, loop):
    connection_data = copy.copy(mysql_server['conn_params'])
    connection_data['user'] = 'nopass_sha256'
    connection_data['password'] = None

    async with create_pool(**connection_data,
                           loop=loop) as pool:
        async with pool.get() as conn:
            # User doesnt have any permissions to look at DBs
            # But as 8.0 will default to caching_sha2_password
            assert conn._auth_plugin_used == 'sha256_password'


@pytest.mark.mysql_verison('8.0')
@pytest.mark.run_loop
async def test_sha256_pw(mysql_server, loop):
    connection_data = copy.copy(mysql_server['conn_params'])
    connection_data['user'] = 'user_sha256'
    connection_data['password'] = 'pass_sha256'

    async with create_pool(**connection_data,
                           loop=loop) as pool:
        async with pool.get() as conn:
            # User doesnt have any permissions to look at DBs
            # But as 8.0 will default to caching_sha2_password
            assert conn._auth_plugin_used == 'sha256_password'


@pytest.mark.mysql_verison('8.0')
@pytest.mark.run_loop
async def test_cached_sha256_nopw(mysql_server, loop):
    connection_data = copy.copy(mysql_server['conn_params'])
    connection_data['user'] = 'nopass_caching_sha2'
    connection_data['password'] = None

    async with create_pool(**connection_data,
                           loop=loop) as pool:
        async with pool.get() as conn:
            # User doesnt have any permissions to look at DBs
            # But as 8.0 will default to caching_sha2_password
            assert conn._auth_plugin_used == 'caching_sha2_password'


@pytest.mark.mysql_verison('8.0')
@pytest.mark.run_loop
async def test_cached_sha256_pw(mysql_server, loop):
    connection_data = copy.copy(mysql_server['conn_params'])
    connection_data['user'] = 'user_caching_sha2'
    connection_data['password'] = 'pass_caching_sha2'

    async with create_pool(**connection_data,
                           loop=loop) as pool:
        async with pool.get() as conn:
            # User doesnt have any permissions to look at DBs
            # But as 8.0 will default to caching_sha2_password
            assert conn._auth_plugin_used == 'caching_sha2_password'



# ported from:
# https://github.com/aio-libs/aiopg/blob/master/aiopg/sa/connection.py
import weakref

from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.dml import UpdateBase
from sqlalchemy.sql.ddl import DDLElement

from . import exc
from .result import create_result_proxy
from .transaction import (RootTransaction, Transaction,
                          NestedTransaction, TwoPhaseTransaction)
from ..utils import _TransactionContextManager, _SAConnectionContextManager


def noop(k):
    return k


class SAConnection:

    def __init__(self, connection, engine, compiled_cache=None):
        self._connection = connection
        self._transaction = None
        self._savepoint_seq = 0
        self._weak_results = weakref.WeakSet()
        self._engine = engine
        self._dialect = engine.dialect
        self._compiled_cache = compiled_cache

    def execute(self, query, *multiparams, **params):
        """Executes a SQL query with optional parameters.

        query - a SQL query string or any sqlalchemy expression.

        *multiparams/**params - represent bound parameter values to be
        used in the execution.  Typically, the format is a dictionary
        passed to *multiparams:

            await conn.execute(
                table.insert(),
                {"id":1, "value":"v1"},
            )

        ...or individual key/values interpreted by **params::

            await conn.execute(
                table.insert(), id=1, value="v1"
            )

        In the case that a plain SQL string is passed, a tuple or
        individual values in \*multiparams may be passed::

            await conn.execute(
                "INSERT INTO table (id, value) VALUES (%d, %s)",
                (1, "v1")
            )

            await conn.execute(
                "INSERT INTO table (id, value) VALUES (%s, %s)",
                1, "v1"
            )

        Returns ResultProxy instance with results of SQL query
        execution.

        """
        coro = self._execute(query, *multiparams, **params)
        return _SAConnectionContextManager(coro)

    def _base_params(self, query, dp, compiled, is_update):
        """
        handle params
        """
        if dp and isinstance(dp, (list, tuple)):
            if is_update:
                dp = {c.key: pval for c, pval in zip(query.table.c, dp)}
            else:
                raise exc.ArgumentError(
                    "Don't mix sqlalchemy SELECT "
                    "clause with positional "
                    "parameters"
                )
        compiled_params = compiled.construct_params(dp)
        processors = compiled._bind_processors
        params = [{
            key: processors.get(key, noop)(compiled_params[key])
            for key in compiled_params
        }]
        post_processed_params = self._dialect.execute_sequence_format(params)
        return post_processed_params[0]

    async def _executemany(self, query, dps, cursor):
        """
        executemany
        """
        result_map = None
        if isinstance(query, str):
            await cursor.executemany(query, dps)
        elif isinstance(query, DDLElement):
            raise exc.ArgumentError(
                    "Don't mix sqlalchemy DDL clause "
                    "and execution with parameters"
                )
        elif isinstance(query, ClauseElement):
            compiled = query.compile(dialect=self._dialect)
            params = []
            is_update = isinstance(query, UpdateBase)
            for dp in dps:
                params.append(
                    self._base_params(
                        query,
                        dp,
                        compiled,
                        is_update,
                    )
                )
            await cursor.executemany(str(compiled), params)
            result_map = compiled._result_columns
        else:
            raise exc.ArgumentError(
                "sql statement should be str or "
                "SQLAlchemy data "
                "selection/modification clause"
            )
        ret = await create_result_proxy(
            self,
            cursor,
            self._dialect,
            result_map
        )
        self._weak_results.add(ret)
        return ret

    async def _execute(self, query, *multiparams, **params):
        cursor = await self._connection.cursor()
        dp = _distill_params(multiparams, params)
        if len(dp) > 1:
            return await self._executemany(query, dp, cursor)
        elif dp:
            dp = dp[0]

        result_map = None
        if isinstance(query, str):
            await cursor.execute(query, dp or None)
        elif isinstance(query, ClauseElement):
            if self._compiled_cache is not None:
                key = query
                compiled = self._compiled_cache.get(key)
                if not compiled:
                    compiled = query.compile(dialect=self._dialect)
                    if dp and dp.keys() == compiled.params.keys() \
                            or not (dp or compiled.params):
                        # we only want queries with bound params in cache
                        self._compiled_cache[key] = compiled
            else:
                compiled = query.compile(dialect=self._dialect)

            if not isinstance(query, DDLElement):
                post_processed_params = self._base_params(
                    query,
                    dp,
                    compiled,
                    isinstance(query, UpdateBase)
                )
                result_map = compiled._result_columns
            else:
                if dp:
                    raise exc.ArgumentError("Don't mix sqlalchemy DDL clause "
                                            "and execution with parameters")
                post_processed_params = compiled.construct_params()
                result_map = None
            await cursor.execute(str(compiled), post_processed_params)
        else:
            raise exc.ArgumentError("sql statement should be str or "
                                    "SQLAlchemy data "
                                    "selection/modification clause")

        ret = await create_result_proxy(
            self, cursor, self._dialect, result_map
        )
        self._weak_results.add(ret)
        return ret

    async def scalar(self, query, *multiparams, **params):
        """Executes a SQL query and returns a scalar value."""
        res = await self.execute(query, *multiparams, **params)
        return (await res.scalar())

    @property
    def closed(self):
        """The readonly property that returns True if connections is closed."""
        return self._connection is None or self._connection.closed

    @property
    def connection(self):
        return self._connection

    def begin(self):
        """Begin a transaction and return a transaction handle.

        The returned object is an instance of Transaction.  This
        object represents the "scope" of the transaction, which
        completes when either the .rollback or .commit method is
        called.

        Nested calls to .begin on the same SAConnection instance will
        return new Transaction objects that represent an emulated
        transaction within the scope of the enclosing transaction,
        that is::

            trans = await conn.begin()   # outermost transaction
            trans2 = await conn.begin()  # "nested"
            await trans2.commit()        # does nothing
            await trans.commit()         # actually commits

        Calls to .commit only have an effect when invoked via the
        outermost Transaction object, though the .rollback method of
        any of the Transaction objects will roll back the transaction.

        See also:
          .begin_nested - use a SAVEPOINT
          .begin_twophase - use a two phase/XA transaction

        """
        coro = self._begin()
        return _TransactionContextManager(coro)

    async def _begin(self):
        if self._transaction is None:
            self._transaction = RootTransaction(self)
            await self._begin_impl()
            return self._transaction
        else:
            return Transaction(self, self._transaction)

    async def _begin_impl(self):
        cur = await self._connection.cursor()
        try:
            await cur.execute('BEGIN')
        finally:
            await cur.close()

    async def _commit_impl(self):
        cur = await self._connection.cursor()
        try:
            await cur.execute('COMMIT')
        finally:
            await cur.close()
            self._transaction = None

    async def _rollback_impl(self):
        cur = await self._connection.cursor()
        try:
            await cur.execute('ROLLBACK')
        finally:
            await cur.close()
            self._transaction = None

    async def begin_nested(self):
        """Begin a nested transaction and return a transaction handle.

        The returned object is an instance of :class:`.NestedTransaction`.

        Nested transactions require SAVEPOINT support in the
        underlying database.  Any transaction in the hierarchy may
        .commit() and .rollback(), however the outermost transaction
        still controls the overall .commit() or .rollback() of the
        transaction of a whole.
        """
        if self._transaction is None:
            self._transaction = RootTransaction(self)
            await self._begin_impl()
        else:
            self._transaction = NestedTransaction(self, self._transaction)
            self._transaction._savepoint = await self._savepoint_impl()
        return self._transaction

    async def _savepoint_impl(self, name=None):
        self._savepoint_seq += 1
        name = 'aiomysql_sa_savepoint_%s' % self._savepoint_seq

        cur = await self._connection.cursor()
        try:
            await cur.execute('SAVEPOINT ' + name)
            return name
        finally:
            await cur.close()

    async def _rollback_to_savepoint_impl(self, name, parent):
        cur = await self._connection.cursor()
        try:
            await cur.execute('ROLLBACK TO SAVEPOINT ' + name)
        finally:
            await cur.close()
        self._transaction = parent

    async def _release_savepoint_impl(self, name, parent):
        cur = await self._connection.cursor()
        try:
            await cur.execute('RELEASE SAVEPOINT ' + name)
        finally:
            await cur.close()
        self._transaction = parent

    async def begin_twophase(self, xid=None):
        """Begin a two-phase or XA transaction and return a transaction
        handle.

        The returned object is an instance of
        TwoPhaseTransaction, which in addition to the
        methods provided by Transaction, also provides a
        TwoPhaseTransaction.prepare() method.

        xid - the two phase transaction id.  If not supplied, a
        random id will be generated.
        """

        if self._transaction is not None:
            raise exc.InvalidRequestError(
                "Cannot start a two phase transaction when a transaction "
                "is already in progress.")
        if xid is None:
            xid = self._dialect.create_xid()
        self._transaction = TwoPhaseTransaction(self, xid)
        await self.execute("XA START %s", xid)
        return self._transaction

    async def _prepare_twophase_impl(self, xid):
        await self.execute("XA END '%s'" % xid)
        await self.execute("XA PREPARE '%s'" % xid)

    async def recover_twophase(self):
        """Return a list of prepared twophase transaction ids."""
        result = await self.execute("XA RECOVER;")
        return [row[0] for row in result]

    async def rollback_prepared(self, xid, *, is_prepared=True):
        """Rollback prepared twophase transaction."""
        if not is_prepared:
            await self.execute("XA END '%s'" % xid)
        await self.execute("XA ROLLBACK '%s'" % xid)

    async def commit_prepared(self, xid, *, is_prepared=True):
        """Commit prepared twophase transaction."""
        if not is_prepared:
            await self.execute("XA END '%s'" % xid)
        await self.execute("XA COMMIT '%s'" % xid)

    @property
    def in_transaction(self):
        """Return True if a transaction is in progress."""
        return self._transaction is not None and self._transaction.is_active

    async def close(self):
        """Close this SAConnection.

        This results in a release of the underlying database
        resources, that is, the underlying connection referenced
        internally. The underlying connection is typically restored
        back to the connection-holding Pool referenced by the Engine
        that produced this SAConnection. Any transactional state
        present on the underlying connection is also unconditionally
        released via calling Transaction.rollback() method.

        After .close() is called, the SAConnection is permanently in a
        closed state, and will allow no further operations.
        """
        if self._connection is None:
            return

        if self._transaction is not None:
            await self._transaction.rollback()
            self._transaction = None
        # don't close underlying connection, it can be reused by pool
        # conn.close()
        self._engine.release(self)
        self._connection = None
        self._engine = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


def _distill_params(multiparams, params):
    """Given arguments from the calling form *multiparams, **params,
    return a list of bind parameter structures, usually a list of
    dictionaries.

    In the case of 'raw' execution which accepts positional parameters,
    it may be a list of tuples or lists.

    """

    if not multiparams:
        if params:
            return [params]
        else:
            return []
    elif len(multiparams) == 1:
        zero = multiparams[0]
        if isinstance(zero, (list, tuple)):
            if not zero or hasattr(zero[0], '__iter__') and \
                    not hasattr(zero[0], 'strip'):
                # execute(stmt, [{}, {}, {}, ...])
                # execute(stmt, [(), (), (), ...])
                return zero
            else:
                # execute(stmt, ("value", "value"))
                return [zero]
        elif hasattr(zero, 'keys'):
            # execute(stmt, {"key":"value"})
            return [zero]
        else:
            # execute(stmt, "value")
            return [[zero]]
    else:
        if (hasattr(multiparams[0], '__iter__') and
                not hasattr(multiparams[0], 'strip')):
            return multiparams
        else:
            return [multiparams]
